/*
Copyright 2022 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package validator

import (
	"net"
	"net/mail"
	"reflect"
	"strings"
	"unicode/utf8"
)

type Validator struct {
	FieldErrors map[string]string
}

func (v *Validator) Valid() bool {
	return len(v.FieldErrors) == 0
}

func (v *Validator) AddFieldError(key, message string) {
	if v.FieldErrors == nil {
		v.FieldErrors = make(map[string]string)
	}

	if _, exists := v.FieldErrors[key]; !exists {
		v.FieldErrors[key] = message
	}
}

func (v *Validator) CheckField(ok bool, key, message string) {
	if !ok {
		v.AddFieldError(key, message)
	}
}

func NotBlank(value string) bool {
	return strings.TrimSpace(value) != ""
}

func NotNil(value any) bool {
	val := reflect.ValueOf(value)

	return !val.IsNil()
}

func MaxChars(value string, n int) bool {
	return utf8.RuneCountInString(strings.TrimSpace(value)) <= n
}

func MinChars(value string, n int) bool {
	return utf8.RuneCountInString(strings.TrimSpace(value)) >= n
}

func PermittedInt(value int, permittedValues ...int) bool {
	for i := range permittedValues {
		if value == permittedValues[i] {
			return true
		}
	}

	return false
}

func PermittedString(value string, permittedValues ...string) bool {
	for i := range permittedValues {
		if value == permittedValues[i] {
			return true
		}
	}

	return false
}

func PermittedStringSet(value []string, permittedValues []string) bool {
	if len(value) == 0 {
		return true
	}

	valueMap := make(map[string]struct{}, len(permittedValues))
	for _, v := range permittedValues {
		valueMap[v] = struct{}{}
	}

	for j := range value {
		if _, ok := valueMap[value[j]]; !ok {
			return false
		}
	}

	return true
}

func IsEmail(value string) bool {
	addr, err := mail.ParseAddress(value)
	if err != nil {
		return false
	}

	parts := strings.SplitN(addr.Address, "@", 2)

	mxs, err := net.LookupMX(parts[1])
	if err != nil || len(mxs) < 1 {
		return false
	}

	return true
}
